from setuptools import setup

setup(
    name="monotonicnetworks",
    version="1.5.0",
    packages=["monotonicnetworks"],
    url="",
    license="MIT",
    author="Ouail Kitouni, Niklas Nolte",
    author_email="kitouni@mit.edu, nolte@meta.com",
    description=(
        "Pytorch implementation of constrained weight operator norms for robustness "
        + "and monotonicity in neural networks."
    ),
)
